/*****************************************************************************
 
 i2clcd.h - LCD over I2C library 
                Designed for HD44870 based LCDs with I2C expander PCF8574X
                on Atmels AVR MCUs
 
 Copyright (C) 2006 Nico Eichelmann and Thomas Eichelmann
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
 You can contact the authors at info@computerheld.de
 
*****************************************************************************/
 #include "../config.h"
#ifndef _I2CLCD_H
#define _I2CLCD_H
#include <util/delay.h>
#include <stdbool.h>
#include <stdint.h> 

#include "i2c.h"

//--System-Configuration-Settings------------------------------------------------------------------------------------


#define wait1us _delay_loop_1((F_CPU * 0.000001) / 3)   
#define wait1ms _delay_loop_2((F_CPU * 0.001) / 4)      

//-------------------------------------------------------------------------------------------------------------------


//--Display-Configuration-Settings-----------------------------------------------------------------------------------

#define LCD_I2C_DEVICE                  0x40    
#define LCD_LINES                       4       
#define LCD_ROWS                        20      
#define LCD_LINE1                       0x00    
#define LCD_LINE2                       0x40    
#define LCD_LINE3                       0x10    
#define LCD_LINE4                       0x50    


#define ON	0
#define OFF	1 


//-------------------------------------------------------------------------------------------------------------------

//--The-following-definitions-are-corresponding-to-the-PIN-Assignment-(see-above)------------------------------------

#define LCD_D4_PIN                      0       
#define LCD_D5_PIN                      1       
#define LCD_D6_PIN                      2       
#define LCD_D7_PIN                      3       
#define LCD_RS_PIN                      4       
#define LCD_RW_PIN                      5       
#define LCD_E_PIN                       6       
#define LCD_BL_PIN                      7      
//-------------------------------------------------------------------------------------------------------------------

#define LCD_D0                          (1 << LCD_D4_PIN)       
#define LCD_D1                          (1 << LCD_D5_PIN)       
#define LCD_D2                          (1 << LCD_D6_PIN)       
#define LCD_D3                          (1 << LCD_D7_PIN)       
#define LCD_D4                          (1 << LCD_D4_PIN)       
#define LCD_D5                          (1 << LCD_D5_PIN)       
#define LCD_D6                          (1 << LCD_D6_PIN)       
#define LCD_D7                          (1 << LCD_D7_PIN)       
#define LCD_RS                          (1 << LCD_RS_PIN)       
#define LCD_RW                          (1 << LCD_RW_PIN)       
#define LCD_BL                          (1 << LCD_BL_PIN)    
#define LCD_E                           (1 << LCD_E_PIN)        


#define LCD_ADDRESS                     0       
#define LCD_DATA                        1       

//-LCD-COMMANDS------------------------------------------------------------------------------------------------------

#define LCD_CLEAR                       0x01    
#define LCD_HOME                        0x02    


#define LCD_ENTRYMODE                   0x04                    
        #define LCD_INCREASE            LCD_ENTRYMODE | 0x02    
        #define LCD_DECREASE            LCD_ENTRYMODE | 0x00    
        #define LCD_DISPLAYSHIFTON      LCD_ENTRYMODE | 0x01    
        #define LCD_DISPLAYSHIFTOFF     LCD_ENTRYMODE | 0x00    


#define LCD_DISPLAYMODE                 0x08                    
        #define LCD_DISPLAYON           LCD_DISPLAYMODE | 0x04  
        #define LCD_DISPLAYOFF          LCD_DISPLAYMODE | 0x00  
        #define LCD_CURSORON            LCD_DISPLAYMODE | 0x02  
        #define LCD_CURSOROFF           LCD_DISPLAYMODE | 0x00  
        #define LCD_BLINKINGON          LCD_DISPLAYMODE | 0x01  
        #define LCD_BLINKINGOFF         LCD_DISPLAYMODE | 0x00 



#define LCD_SHIFTMODE                   0x10                    
        #define LCD_DISPLAYSHIFT        LCD_SHIFTMODE | 0x08    
        #define LCD_CURSORMOVE          LCD_SHIFTMODE | 0x00    
        #define LCD_RIGHT               LCD_SHIFTMODE | 0x04    
        #define LCD_LEFT                LCD_SHIFTMODE | 0x00    


#define LCD_CONFIGURATION               0x20                            
        #define LCD_8BIT                LCD_CONFIGURATION | 0x10        
        #define LCD_4BIT                LCD_CONFIGURATION | 0x00        
        #define LCD_2LINE               LCD_CONFIGURATION | 0x08        
        #define LCD_1LINE               LCD_CONFIGURATION | 0x00        
        #define LCD_5X10                LCD_CONFIGURATION | 0x04        
        #define LCD_5X7                 LCD_CONFIGURATION | 0x00        

//-------------------------------------------------------------------------------------------------------------------

//-FUNCTIONS---------------------------------------------------------------------------------------------------------

void lcd_init(void);                                    //-     Display initialization sequence

void lcd_write_i2c(unsigned char value);                //-     Write data to i2c

void lcd_write(unsigned char value);                    //-     Write byte to display with toggle of enable-bit

bool lcd_gotolr(unsigned char line, unsigned char row); //-     Go to position

void lcd_putchar(unsigned char value);                  //-     Put char to cursor position

bool lcd_putcharlr(unsigned char line, unsigned char row, unsigned char value); //-     Put char to position

void lcd_print(unsigned char *string);                  //-     Print string to cursor position

bool lcd_printlr(unsigned char line, unsigned char row , unsigned char *string);        //-     Print string to position

bool lcd_printlc(unsigned char line, unsigned char row, unsigned char *string);         //-     Print string to position 
                                                                                        //-     (If string is longer than LCD_ROWS
                                                                                        //-     overwrite first chars)

bool lcd_printlrc(unsigned char line, unsigned char row, unsigned char *string);        //-     Print string to position 
                                                                                        //-     (If string is longer than LCD_ROWS
                                                                                        //-     continue in next line)

void lcd_command(unsigned char command);                //-     Issue a command to the display (use the defined commands above)

bool lcd_nextline(void);                                //-     Go to nextline (if next line > LCD_LINES return false)

unsigned char lcd_read_i2c(void);                       //-     Read data from i2c

unsigned char lcd_read(bool mode);                      //-     Read data from display over i2c
                                                        //-     (second nibble corresponds to highbyte or lowbyte)

unsigned char lcd_getbyte(bool mode);                   //-     Read one complete byte over i2c from display

bool lcd_getlr(unsigned char *line, unsigned char*row); //-     Get line and row (target byte for line, target byte for row)

bool lcd_busy(void);                                    //-     Check if busy

void lcd_wait_us(unsigned short us);                    //-     Wait some microseconds

void lcd_wait_ms(unsigned short ms);                    //-     Wait some milliseconds

void lcd_backlight(int bl);			//
//-------------------------------------------------------------------------------------------------------------------
#endif
